package WebGPU::Direct::SupportedFeatures
{
  use v5.30;
  use warnings;
  no warnings qw(experimental::signatures);
  use feature 'signatures';
};

1;
__END__
=pod

=encoding UTF-8

=head1 NAME

WebGPU::Direct::SupportedFeatures

=head2 Methods

=head3 freeMembers

